import random
import argparse
from agt_server.agents.base_agents.rps_agent import RPSAgent
from agt_server.local_games.rps_arena import RPSArena


class TAAgent(RPSAgent):
    def setup(self):
        self.ROCK, self.SCISSORS, self.PAPER = 0, 1, 2
        self.actions = [self.ROCK, self.PAPER]

    def get_action(self):
        return random.choice(self.actions)

    def update(self):
        return None


if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = TAAgent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = RPSArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                TAAgent("TA_Agent_1"),
                TAAgent("TA_Agent_2"),
                TAAgent("TA_Agent_3"),
                TAAgent("TA_Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np

class DumbChicken(LemonadeAgent):
    def setup(self):
        pass

    def get_action(self):
        if (len(self.get_opp1_action_history()) != 0):
            opp1 = np.floor(np.mean(np.array(self.get_opp1_action_history())))
            opp2 = np.floor(np.mean(np.array(self.get_opp2_action_history())))
            opp1_opp = (opp1 + 6) % 12
            opp2_opp = (opp2 + 6) % 12
            mean = (opp1_opp + opp2_opp) / 2
            action = np.floor(mean)
        else:
            action = random.randrange(12)
        return int(action)

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = DumbChicken("Dumb")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = DumbChicken(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                DumbChicken("Agent_1"),
                DumbChicken("Agent_2"),
                DumbChicken("Agent_3"),
                DumbChicken("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.agent import Agent
import json
import pandas as pd
import threading


class CompleteMatrixAgent(Agent):
    def calculate_utils(self, p1_action, p2_action):
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            utils = [0, 0]
        elif p1_action not in self.valid_actions:
            utils = [self.invalid_move_penalty, 0]
        elif p2_action not in self.valid_actions:
            utils = [0, self.invalid_move_penalty]
        else:
            utils = [self.utils[p1_action][p2_action][0],
                     self.utils[p1_action][p2_action][1]]
        return utils

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        self.player_type = resp['player_type']
        if 'all' in resp['permissions']:
            
            self.game_report.game_history['my_action_history'].append(
                resp['my_action'])
            self.game_report.game_history['my_utils_history'].append(
                resp['my_utils'])
            self.game_report.game_history['opp_action_history'].append(
                resp['opp_action'])
            self.game_report.game_history['opp_utils_history'].append(
                resp['opp_utils'])
        else:
            for perm in resp['permissions']:
                self.game_report.game_history[f'{perm}_history'].append(
                    resp[perm])

    def handle_postround_data(self, resp):
        self.global_timeout_count = resp['global_timeout_count']
        self.curr_opps = resp['opp_names']
        self.handle_permissions(resp)

    def play(self):
        data = self.client.recv(1024).decode()
        if data:
            resp = json.loads(data)
            if resp['message'] == 'provide_game_name':
                print(f"We are playing {resp['game_name']}")
                message = {
                    "message": "game_name_recieved",
                }
                self.client.send(json.dumps(message).encode())
                self.restart()
        while True:
            data = self.client.recv(10000).decode()
            if data:
                request = json.loads(data)
                if request['message'] == 'send_preround_data':
                    self.player_type = request['player_type']
                    message = {"message": "preround_data_recieved"}
                    self.client.send(json.dumps(message).encode())
                    continue
                elif request['message'] == 'request_action':
                    self.timeout = False
                    try:
                        timer = threading.Timer(self.response_time, self.timeout_handler)
                        timer.start()
                        action = self.get_action()
                    finally:
                        if self.timeout: 
                            action = -1
                        timer.cancel()

                    try:
                        action = int(action)
                        message = {
                            "message": "provide_action",
                            "action": action, 
                            "timeout": self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                    except:
                        print("Warning: Get Action must return an Integer")
                        message = {
                            "message": "provide_action",
                            "action": -1, 
                            "timeout":self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                elif request['message'] == 'game_end':
                    if request['send_results']:
                        try: 
                            df = pd.read_json(request['results'])
                            if df is not None:
                                print(df)
                        except: 
                            print("Results too large. Please check in with your Lab TA for the results")
                    else:
                        print(request['results'])
                    self.close()
                    break

            data = self.client.recv(10000).decode()
            if data:
                resp = json.loads(data)
                if resp['message'] == 'prepare_next_game':
                    self.print_results()
                    self.restart()
                    message = {"message": "ready_next_game"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'prepare_next_round':
                    self.handle_postround_data(resp)
                    self.update()
                    message = {"message": "ready_next_round"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'disqualified':
                    if resp['disqualification_message']:
                        print(resp['disqualification_message'])
                    self.close()
                    break

    def get_game_report(self): 
        return self.game_report
    
    def get_action_history(self):
        return self.game_report.get_action_history()

    def get_util_history(self):
        return self.game_report.get_util_history()

    def get_opp_action_history(self):
        return self.game_report.get_opp_action_history()

    def get_opp_util_history(self):
        return self.game_report.get_opp_util_history()
    
    def get_last_action(self):
        return self.game_report.get_last_action()

    def get_last_util(self):
        return self.game_report.get_last_util()

    def get_opp_last_action(self):
        return self.game_report.get_opp_last_action()

    def get_opp_last_util(self):
        return self.game_report.get_opp_last_util()


from agt_server.local_games.base import LocalArena
import numpy as np
from collections import defaultdict
import gzip
import json
from datetime import datetime
import pkg_resources
from elommr import EloMMR, Player, TanhTerm, PlayerEvent
import random
from copy import deepcopy
import pandas as pd
import os
import math
from tqdm import tqdm

class LSVMArena(LocalArena):
    def __init__(self, num_cycles_per_player = 3, players=[], timeout=1, handin=False, logging_path = None, save_path = None, elo_path = "local_elo_ranking.json", local_save_path = None, num_rounds = 1000):
        slack = 1000 
        random_number = np.random.exponential(scale=slack / 2)
        random_number_clipped = min(max(0, random_number), slack)
        num_rounds += random_number_clipped
        num_rounds = math.ceil(num_rounds)
        super().__init__(num_rounds, players, timeout, handin, logging_path, save_path)
        self.game_name = "Spectrum Auction - Local Synergy Value Model (LSVM)"
        self.num_cycles_per_player = num_cycles_per_player
        self.elommr = EloMMR(max_history = 10000)
        self.epsilon = 0.1
        self.current_prices = None
        self.min_bids = None
        self.tentative_winners = None
        self.tentative_winners_map = {}
        self.elo_path = elo_path
        self.local_save_path = local_save_path
    
        if not self.handin_mode:
            assert len(self.players) >= 6, "Arena must have at least 6 players"
            player_names = [player.name for player in players]
            assert len(set(player_names)) == len(player_names), "Players must have unique names"

        for idx in range(len(self.players)):
            if self.handin_mode:
                try:
                    player = self.players[idx]
                    self.game_reports[player.name] = {
                        "bid_history": [],
                        "price_history": [],
                        "util_history": [],
                        "winner_history": [],
                        "index": idx,
                        "timeout_count": 0,
                        "global_timeout_count": 0,
                        "disconnected": False, 
                        "elo": self.load_player_from_json(player.name)
                    }
                except:
                    self.players[idx] = None
                    continue
            else:
                player = self.players[idx]
                self.game_reports[player.name] = {
                    "bid_history": [],
                    "price_history": [],
                    "util_history": [],
                    "winner_history": [],
                    "index": idx,
                    "global_timeout_count": 0, 
                    "elo": self.load_player_from_json(player.name)
                }
        

        self.results = []
        self.game_num = 1
        
    def save_player_to_json(self, name, player):
        try:
            with open(self.elo_path, 'r') as f:
                data = json.load(f)
        except FileNotFoundError:
            data = {}
        
        data[name] = {
            "_normal_factor": {"mu": player._normal_factor.mu, "sig": player._normal_factor.sig},
            "_logistic_factors": [{"mu": tht.mu, "w_arg": tht.w_arg, "w_out": tht.w_out} for tht in player._logistic_factors],
            "event_history": [{"mu": ev.mu, "sig": ev.sig, "perf_score": ev.perf_score, "place": ev.place} for ev in player.event_history],
            "approx_posterior": {"mu": player.approx_posterior.mu, "sig": player.approx_posterior.sig},
            "update_time": player.update_time,
            "delta_time": player.delta_time
        }
        
        with open(self.elo_path, 'w') as f:
            json.dump(data, f, indent=4)


    def load_player_from_json(self,name):
        with open(self.elo_path, 'r') as f:
            data = json.load(f)
        
        player = Player()
        if name in data:
            player_dict = data[name]
            player._normal_factor.mu = player_dict["_normal_factor"]["mu"]
            player._normal_factor.sig = player_dict["_normal_factor"]["sig"]
            player._logistic_factors = [TanhTerm(tht['mu'], tht['w_arg'], tht['w_out']) for tht in player_dict["_logistic_factors"]]
            player.event_history = [PlayerEvent(ev['mu'], ev['sig'], ev['perf_score'], ev['place']) for ev in player_dict["event_history"]]
            player.approx_posterior.mu = player_dict["approx_posterior"]["mu"]
            player.approx_posterior.sig = player_dict["approx_posterior"]["sig"]
            player.update_time = player_dict["update_time"]
            player.delta_time = player_dict["delta_time"]
        return player
    

    def reset_game_reports(self):
        for player in self.players:
            if player is not None: 
                if self.handin_mode:
                    try:
                        self.game_reports[player.name]["bid_history"] = []
                        self.game_reports[player.name]["price_history"] = []
                        self.game_reports[player.name]["util_history"] = []
                        self.game_reports[player.name]["winner_history"] = []
                    except:
                        continue
                else:
                    self.game_reports[player.name]["bid_history"] = []
                    self.game_reports[player.name]["price_history"] = []
                    self.game_reports[player.name]["util_history"] = []
                    self.game_reports[player.name]["winner_history"] = []

    def _find_matches(self, player_name):
        assert player_name in self.game_reports, "Cannot match up a player that does not exist in the program."
        if len(self.game_reports[player_name]['elo'].event_history) > 0:
            target_rating = self.game_reports[player_name]['elo'].event_history[-1].mu
        else:
            target_rating = 400

        rating_diffs = {}
        for name in self.game_reports:
            if name != player_name and (not self.handin_mode or not self.game_reports[name]['disconnected']):
                if len(self.game_reports[name]['elo'].event_history) > 0:
                    rating_diffs[name] = abs(self.game_reports[name]['elo'].event_history[-1].mu - target_rating)
                else:
                    rating_diffs[name] = 400 - target_rating

        # TODO: See if ELO is fixable/integratable later
        if False:
            closest_matches = sorted(rating_diffs, key=rating_diffs.get)[:5]
            return set(closest_matches)
        else:
            all_possible_matches = list(rating_diffs.keys())
            random_matches = random.sample(all_possible_matches, min(5, len(all_possible_matches)))
            return set(random_matches)
    
    
    def run(self):
        """
        Simulates an LSVM auction with all of the players in self.players. 
        For each player, they are matched with the 5 other players closest in ELO to that player, 
        Then for the number of cycles each player plays, each player will cycle through being the national bidder once in a match 
        In each match the other regional bidders will be assigned one of the regional goods uniform at random. 
        Note that the regional bidders can be assigned the same regional good. 
        """
        for player in self.players:
            if player is not None: 
                for _ in range(self.num_cycles_per_player):
                    other_players = [self.players[self.game_reports[op_name]['index']] for op_name in self._find_matches(player.name)]
                    curr_players = other_players + [player]
                    for i in range(len(curr_players)): 
                        self.current_round = 0
                        shape = curr_players[i].get_shape()
                        self.current_prices = np.zeros(shape)
                        self.min_bids = np.zeros(shape)
                        self.tentative_winners = np.empty(shape, dtype=object)
                        self.tentative_winners_map = {}
                        curr_players[i]._is_national_bidder = True 
                        curr_players[i].valuations = np.random.uniform(3, 9, shape)
                        curr_players[i]._current_round = self.current_round
                        for j in range(len(curr_players)): 
                            if i != j: 
                                curr_players[j]._is_national_bidder = False
                                curr_players[j].regional_good = np.random.choice(list("ABCDEFGHIJKLMNOPQR"))
                                curr_players[j].valuations = curr_players[j].proximity(np.random.uniform(3, 20, shape))
                                curr_players[j]._current_round = self.current_round
                        if self.handin_mode: 
                            if any([p is None or self.game_reports[p.name]['disconnected'] for p in curr_players]):
                                continue
                            else: 
                                for p in curr_players: 
                                    try:
                                        self.run_func_w_time(
                                            p.restart, self.timeout, p.name)
                                    except:
                                        self.game_reports[p.name]['disconnected'] = True
                                        continue
                            try:
                                self.run_game(curr_players)
                            except:
                                self.reset_game_reports()
                                continue
                            
                            for p in curr_players: 
                                if p is None or self.game_reports[p.name]['disconnected']: 
                                    continue 
                                else: 
                                    try:
                                        self.run_func_w_time(
                                            p.teardown, self.timeout, p.name)
                                    except:
                                        self.game_reports[p.name]['disconnected'] = True
                                        continue
                        else: 
                            for p in curr_players: 
                                self.run_func_w_time(p.restart, self.timeout, p.name)
                            self.run_game(curr_players)
        results = self.summarize_results()
        return results

    @staticmethod
    def prune_valid_bids(player, my_bids):
        """
        Check if my_bids is a valid bid bundle and prunes the bids that are not valid.

        :param my_bids: Dictionary mapping goods to bid values.
        :return: True if the bid bundle is valid, False otherwise.
        """
        
        if not isinstance(my_bids, dict):
            return {}
        
        valid_bids = {}
        for good, bid in my_bids.items():
            if bid is None:
                continue

            if good not in player._goods_to_index or bid < player.min_bids[player._goods_to_index[good]]:
                continue

            price_history = player.game_report.game_history['price_history']
            bid_history = player.game_report.game_history['bid_history']
            revealed_preference = True
            my_bids_arr = player.map_to_ndarray(my_bids)
            if len(price_history) > 0 and len(bid_history) > 0: 
                past_prices = price_history[-1]
                past_bids = bid_history[-1]
                price_diff = player.current_prices - past_prices
                bid_diff = my_bids_arr - past_bids
                switch_cost = np.dot(price_diff.reshape(-1), bid_diff.reshape(-1))
                if switch_cost > 0:
                    revealed_preference = False 
                    break 
            
            if revealed_preference: 
                valid_bids[good] = bid
        return valid_bids 
    
    def run_helper(self, curr_players):
        has_incremental_goods = True
        rounds_so_far = 0
        pbar = tqdm(total = self.num_rounds, desc="Running Round", leave=False)
        
        if self.handin_mode:
            for p in curr_players: 
                try:
                    self.run_func_w_time(
                        p.setup, self.timeout, p.name)
                except:
                    self.game_reports[p.name]['disconnected'] = True
        else: 
            for p in curr_players:
                self.run_func_w_time(p.setup, self.timeout, p.name)
        
        while has_incremental_goods and rounds_so_far <= self.num_rounds: 
            rounds_so_far += 1
            pbar.update(1)
            has_incremental_goods = False
            if self.handin_mode:
                if any([self.game_reports[p.name]['disconnected'] for p in curr_players]):
                    return
                bids_this_round = defaultdict(lambda: [])
                for p in curr_players: 
                    p.current_prices = self.current_prices
                    p.min_bids = self.current_prices + self.epsilon
                    if self.game_reports[p.name]['timeout_count'] < self.timeout_tolerance: 
                        try:
                            bids = self.run_func_w_time(
                                p.get_bids, self.timeout, p.name, {})
                            pruned_bids = LSVMArena.prune_valid_bids(p, bids)
                            for good in pruned_bids: 
                                bids_this_round[good].append((p.name, pruned_bids[good]))
                        except:
                            self.game_reports[p.name]['disconnected'] = True
                    else:
                        self.game_reports[p.name]['disconnected'] = True
                
                    self.game_reports[p.name]['bid_history'].append(bids)
                    p.game_report.game_history['bid_history_map'].append(bids)
                    p.game_report.game_history['bid_history'].append(p.map_to_ndarray(bids))
                    
                for p in curr_players: 
                    p.tentative_allocation = p.tentative_allocation - set(bids_this_round.keys())
                for good in bids_this_round: 
                    if len(bids_this_round[good]) > 0: 
                        bids = bids_this_round[good]
                        if len(bids) > 1: 
                            has_incremental_goods = True
                        unique_bid_values = set([bid for _, bid in bids])
                        sorted_bid_values = sorted(list(unique_bid_values), reverse=True)
                        highest_bid = sorted_bid_values[0]
                        winners = [name for name, bid in bids if bid == highest_bid]
                        winner_name = random.choice(winners)
                        winner = self.players[self.game_reports[winner_name]['index']]
                        winner_idx = winner._goods_to_index[good]
                        winner.tentative_allocation.add(good)
                        self.tentative_winners[winner_idx] = winner_name
                        self.tentative_winners_map[good] = winner_name
                        if len(sorted_bid_values) <= 1: 
                            self.current_prices[winner_idx] += self.epsilon
                        else: 
                            self.current_prices[winner_idx] = sorted_bid_values[1]
                
            else:
                bids_this_round = defaultdict(lambda: [])                
                for p in curr_players: 
                    p.current_prices = self.current_prices
                    p.min_bids = self.current_prices + self.epsilon
                    bids = self.run_func_w_time(
                        p.get_bids, self.timeout, p.name, {})
                    pruned_bids = LSVMArena.prune_valid_bids(p, bids)
                    for good in pruned_bids: 
                        bids_this_round[good].append((p.name, pruned_bids[good]))
                    self.game_reports[p.name]['bid_history'].append(bids)
                    p.game_report.game_history['bid_history_map'].append(bids)
                    p.game_report.game_history['bid_history'].append(p.map_to_ndarray(bids))
                    
                for p in curr_players: 
                    p.tentative_allocation = p.tentative_allocation - set(bids_this_round.keys())
                for good in bids_this_round: 
                    if len(bids_this_round[good]) > 0: 
                        bids = bids_this_round[good]
                        if len(bids) > 1: 
                            has_incremental_goods = True
                        unique_bid_values = set([bid for _, bid in bids])
                        sorted_bid_values = sorted(list(unique_bid_values), reverse=True)
                        highest_bid = sorted_bid_values[0]
                        winners = [name for name, bid in bids if bid == highest_bid]
                        winner = random.choice(winners)
                        winner_name = random.choice(winners)
                        winner = self.players[self.game_reports[winner_name]['index']]
                        winner_idx = winner._goods_to_index[good]
                        winner.tentative_allocation.add(good)
                        self.tentative_winners[winner_idx] = winner_name
                        self.tentative_winners_map[good] = winner_name
                        if len(sorted_bid_values) <= 1: 
                            self.current_prices[winner_idx] += self.epsilon
                        else: 
                            self.current_prices[winner_idx] = sorted_bid_values[1]

            for p in curr_players: 
                prices_map = p.ndarray_to_map(p.current_prices)
                self.game_reports[p.name]['price_history'].append(prices_map)
                self.game_reports[p.name]['util_history'].append(p.calc_total_utility())
                self.game_reports[p.name]['winner_history'].append(self.tentative_winners_map)
                
                p.game_report.game_history['price_history'].append(p.current_prices)
                p.game_report.game_history['price_history_map'].append(prices_map)
                p.game_report.game_history['my_utils_history'].append(self.game_reports[p.name]['util_history'])
                p.game_report.game_history['winner_history'].append(self.tentative_winners)
                p.game_report.game_history['winner_history_map'].append(self.tentative_winners_map)

            self.current_round += 1
            for p in curr_players: 
                p._current_round = self.current_round
            
            if self.handin_mode:
                for p in curr_players:
                    try:
                        self.run_func_w_time(p.update, self.timeout, p.name)
                    except:
                        self.game_reports[p.name]['disconnected'] = True
            else:
                for p in curr_players:
                    self.run_func_w_time(p.update, self.timeout, p.name)
                    self.run_func_w_time(p.update, self.timeout, p.name)
                    self.run_func_w_time(p.update, self.timeout, p.name)
        pbar.close()
                        
    def run_game(self, curr_players):
        curr_player_names = [p.name for p in curr_players]
        curr_player_names_modified = [p.name + " (National)" if p.is_national_bidder() else p.name for p in curr_players]
        
        versus_str = f"Auction {self.game_num}: " + " VS ".join(curr_player_names_modified)
        print(f"{versus_str}")
        if self.handin_mode: 
            with open(self.logging_path, 'a') as file:
                file.write(f"\n{versus_str}\n")
        
        self.run_helper(curr_players)
        
        total_u = []
        for p in curr_players: 
            total_u.append(self.print_results(p))    
            
        player_utilities = sorted(zip(curr_player_names, total_u), key=lambda x: x[1], reverse=True)

        standings = []
        tie_start_index = 0
        
        for i in tqdm(range(len(player_utilities)), desc = "Calculating Standings", leave=False):
            if i == len(player_utilities) - 1 or player_utilities[i][1] != player_utilities[i + 1][1]:
                for j in range(tie_start_index, i + 1):
                    standings.append((player_utilities[j][0], tie_start_index, i))
                tie_start_index = i + 1 
                
        winner = player_utilities[0][0]       
        standings = [(self.game_reports[name]['elo'], start, end) for name, start, end in standings]
        self.elommr.round_update(standings)
        
        for p_name in curr_player_names: 
            self.save_player_to_json(p_name, self.game_reports[p_name]['elo'])
        
        sorted_zip = sorted(zip(curr_players, total_u), key=lambda x: x[0].name)
        sorted_players, sorted_total_u = zip(*sorted_zip)
        sorted_names = [p.name for p in sorted_players]
        self.results.append(sorted_names + list(sorted_total_u) + [winner])
        
        if self.local_save_path != None:
            now = datetime.now()
            timestamp_str = now.strftime("%Y-%m-%d_%H-%M-%S")
            file_name = f"{timestamp_str}.json.gz"

            final_game_reports = deepcopy(self.game_reports)
            for player_name in final_game_reports: 
                player = self.players[final_game_reports[player_name]['index']]
                final_game_reports[player_name]['is_national_bidder'] = player.is_national_bidder()
                if player.valuations is not None:
                    final_game_reports[player_name]['valuations'] = player.ndarray_to_map(player.valuations)
                else: 
                    final_game_reports[player_name]['valuations'] = None
                if len(final_game_reports[player_name]['elo'].event_history) > 0:
                    final_game_reports[player_name]['elo'] = final_game_reports[player_name]['elo'].event_history[-1].display_rating()
                else: 
                    final_game_reports[player_name]['elo'] = f"{400} ± {0}"
                final_game_reports[player_name]['regional_good'] = player.get_regional_good()
            if self.handin_mode: 
                #TODO: Somehow interface with google drive to dump this file to a folder or shared drive in google.
                pass 
            else:    
                with gzip.open(f"{self.local_save_path}/{file_name}", 'wt', encoding='UTF-8') as f:
                    json.dump(final_game_reports, f, indent=4)
        
        self.game_num += 1                 
        self.reset_game_reports()

    def print_results(self, p):
        if len(self.game_reports[p.name]['util_history']) > 0:
            final_util = self.game_reports[p.name]['util_history'][-1]
        else: 
            final_util = 0
        if len(p.tentative_allocation) > 0:
            print_smt = f"{p.name} won"
            print_smt += ",".join([f" Good {good} at Price {p.current_prices[p._goods_to_index[good]]:.2f}" for good in p.tentative_allocation])
            print_smt += "."
            if self.handin_mode:
                with open(self.logging_path, 'a') as file:
                    file.write(f"{print_smt}\n")
            else: 
                print(f"{print_smt}")
        
        if not self.handin_mode:
            if self.game_reports[p.name]['global_timeout_count'] > 0:
                print(
                    f"{p.name} timed out {self.game_reports[p.name]['global_timeout_count']} times")
                self.game_reports[p.name]['global_timeout_count'] = 0
        
        if self.handin_mode:
            with open(self.logging_path, 'a') as file:
                    file.write(f"{p.name} got a final utility of {final_util}\n")
        else: 
            print(f"{p.name} got a final utility of {final_util}")
        return final_util

    def summarize_results(self):
        df = pd.DataFrame(self.results)
        df.columns = ['Agent 1', 'Agent 2', 'Agent 3', 'Agent 4', 'Agent 5', 'Agent 6',
                      'A1 Score', 'A2 Score', 'A3 Score', 'A4 Score', 'A5 Score', 'A6 Score', 
                      'Winner']
        df = df.groupby(['Agent 1', 'Agent 2', 'Agent 3', 'Agent 4', 'Agent 5', 'Agent 6'])[['A1 Score', 'A2 Score', 'A3 Score', 'A4 Score', 'A5 Score', 'A6 Score']].mean().reset_index()

        if self.handin_mode: 
            with open(self.logging_path, 'a') as file:
                file.write(f"Extended Results: \n {df}")
        else: 
            print(f"Extended Results: \n {df}")

        total_util_dict = defaultdict(lambda: [0, 0])
        for res in self.results:
            player_names = res[:6]
            utils = res[6:12]
            for name, util in zip(player_names, utils):
                total_util_dict[name][0] += util
                total_util_dict[name][1] += 1      

        res_summary = []
        for key, value in total_util_dict.items():
            if len( self.game_reports[key]['elo'].event_history) > 0:
                elo =  self.game_reports[key]['elo'].event_history[-1].mu
            else: 
                elo = 400
            if self.handin_mode and self.game_reports[key]['disconnected']:
                res_summary.append([key, float('-inf'), elo])
            elif value[1] > 0:
                res_summary.append(
                    [key, value[0] / value[1], elo])
            else:
                res_summary.append([key, 0, elo])

        sum_df = pd.DataFrame(res_summary)
        sum_df.columns = ['Agent Name', 'Final Score', 'ELO']
        sum_df = sum_df.sort_values('ELO', ascending=False)
        sum_df = sum_df[sum_df['Final Score'] != float('-inf')]
        
        if not self.handin_mode:
            print(f"Results: \n {sum_df}")
        else: 
            today_date_formatted = datetime.now().strftime('%Y-%m-%d_%H%M')
            final_str = f"{today_date_formatted}\t"
            result_list = [str(item) for group in zip(sum_df['Agent Name'], sum_df['Final Score'], sum_df['ELO']) for item in group]
            final_str += "\t".join(result_list)
            print(final_str)
            with open(pkg_resources.resource_filename('agt_server', self.save_path), 'a') as file:
                file.write(final_str + "\n")
            
        return sum_df



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np

LEARNING_RATE = 0.05
DISCOUNT_FACTOR = 0.90
EXPLORATION_RATE = 0.05


class CircularHotel(LemonadeAgent):
    def setup(self):
        self.num_states = range(144)
        self.starting_val = 0
        self.q_table = {
            s: [self.starting_val for _ in range(12)] for s in self.num_states}
        self.opp_1_acc = None
        self.opp_2_acc = None
        self.start = False
        self.state = None
        self.action = None

    def get_action(self):
        if self.start:
            state = self.opp_1_acc * 12 + self.opp_2_acc
            self.state = state
            try:
                action = np.argmax(self.q_table[state])
            except:
                action = random.randint(0, 11)
            self.action = action
            return action
        else:
            state = random.randint(0, 144)
            self.start = True
            self.state = state
            self.action = random.randint(0, 11)
            return self.action

    def update(self):
        self.opp_1_acc = self.get_opp1_last_action()
        self.opp_2_acc = self.get_opp2_last_action()
        try:
            self.q_table[self.state][self.action] += LEARNING_RATE * (
                self.get_last_util() + DISCOUNT_FACTOR *
                np.max(self.q_table[self.state])
                - self.q_table[self.state][self.action])
        except:
            pass

################### SUBMISSION #####################
agent_submission = CircularHotel("Circular Hotel")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = CircularHotel(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                CircularHotel("Agent_1"),
                CircularHotel("Agent_2"),
                CircularHotel("Agent_3"),
                CircularHotel("Agent_4")
            ]
        )
        arena.run()



from agt_server.server.games.complete_2x2_matrix_game import Complete2by2MatrixGame


class ChickenGame(Complete2by2MatrixGame):
    def __init__(self, num_rounds=1000, player_data=[], player_types=[], permissions_map={}, game_kick_timeout=60, game_name=None, invalid_move_penalty=0, timeout_tolerance = 10):
        super().__init__(num_rounds, player_data,
                         player_types, permissions_map, game_kick_timeout, game_name, invalid_move_penalty, timeout_tolerance)
        self.valid_actions = [0, 1]
        self.utils = [[(0, 0), (-1, 1)],
                      [(1, -1), (-5, -5)]]



import argparse
import math
import random
from agt_server.agents.base_agents.bosii_agent import BOSIIAgent
from agt_server.local_games.bosii_arena import BOSIIArena


class ChippingFish(BOSIIAgent):
    # You can ignore this test, it is just the agent that I implemented for the class competition moved from the trading platform to here
    def setup(self):
        self.COMPROMISE, self.STUBBORN = 0, 1
        self.GOOD_MOOD, self.BAD_MOOD = 0, 1
        self.actions = [self.COMPROMISE, self.STUBBORN]

    def get_action(self):
        discount = 1
        discount_rate = 0.01
        goodMoodC = 0.0
        goodMoodS = 0.0
        goodTotal = 0.0
        badMoodC = 0.0
        badMoodS = 0.0
        badTotal = 0.0
        probC = 0.0
        probS = 0.0

        mood_history = self.get_mood_history()
        action_history = self.get_action_history()
        util_history = self.get_util_history()
        for mood, action, util in zip(mood_history, action_history, util_history):
            if mood == self.GOOD_MOOD:
                goodTotal += discount
                if action == self.COMPROMISE:
                    goodMoodC += discount * util
                elif action == self.STUBBORN:
                    goodMoodS += discount * util
            elif mood == self.BAD_MOOD:
                badTotal += discount
                if action == self.COMPROMISE:
                    badMoodC += discount * util
                elif action == self.STUBBORN:
                    badMoodS += discount * util
            discount *= 1 - discount_rate

        if self.is_row_player():
            moodProb = self.col_player_good_mood_prob()
            if goodTotal > 0 and badTotal > 0:
                expectedC = (moodProb * (goodMoodC / goodTotal)) + \
                    ((1 - moodProb) * (badMoodC / badTotal))
                expectedS = (moodProb * (goodMoodC / goodTotal)) + \
                    ((1 - moodProb) * (badMoodC / badTotal))
                probC = math.exp(expectedC)
                probS = math.exp(expectedS)
        else:
            if self.get_mood() == self.GOOD_MOOD:
                if goodTotal > 0:
                    probC = math.exp(goodMoodC / goodTotal)
                if badTotal > 0:
                    probS = math.exp(badMoodC / badTotal)
            elif self.get_mood() == self.BAD_MOOD:
                if goodTotal > 0:
                    probC = math.exp(goodMoodS / goodTotal)
                if badTotal > 0:
                    probS = math.exp(badMoodS / badTotal)

        if random.random() < probC:
            return self.COMPROMISE
        else:
            return self.STUBBORN

    def update(self):
        return None


################### SUBMISSION #####################
agent_submission = ChippingFish("Chipping Fish")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = ChippingFish(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = BOSIIArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                ChippingFish("Agent_1"),
                ChippingFish("Agent_2"),
                ChippingFish("Agent_3"),
                ChippingFish("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random


class EyesOut(LemonadeAgent):
    def setup(self):
        pass

    def get_action(self):
        try:
            last_opp_1 = self.get_opp1_action_history()[-1]
            last_opp_2 = self.get_opp2_action_history()[-1]

            diff = abs(last_opp_1-last_opp_2)
            k = random.randint(0, 1)
            min_act = min(last_opp_1,last_opp_2)
            max_act = max(last_opp_1,last_opp_2)
            if diff >= 5 and diff <= 7:
                if diff <= 6:
                    action = (max_act + 4) % 12
                if diff == 7:
                    action = (max_act - 3) % 12
            elif diff == 4 or diff == 8:
                if diff == 8:
                    action = (min_act + 4) % 12
                if diff == 4:
                    action = (max_act + 4) % 12
            elif diff <= 3 or diff>=9:
                action = (max_act + 5) % 12
                if k == 1:
                    action = (min_act - 4) % 12
        except:
            action = random.choice([2])
        return action

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = EyesOut("Etch")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = EyesOut(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                EyesOut("Agent_1"),
                EyesOut("Agent_2"),
                EyesOut("Agent_3"),
                EyesOut("Agent_4")
            ]
        )
        arena.run()



import socket
import json
import uuid
from agt_server.agents.base_agents.game_report import GameReport


class Agent:
    def __init__(self, name=None, timestamp = None):
        self.name = name
        self.timestamp = timestamp
        self.client = None
        self.player_type = None
        self.game_report = GameReport()
        self.game_num = 1
        self.global_timeout_count = 0
        self.curr_opps = []
        self.setup()

    def respond_to_request(self, key, value):
        data = self.client.recv(1024).decode()
        if data:
            request = json.loads(data)
            if request['message'] == f'request_{key}':
                message = {
                    "message": f"provide_{key}",
                    f"{key}": value
                }
                self.client.send(json.dumps(message).encode())

    def setup(self):
        raise NotImplementedError

    def restart(self):
        self.game_report = GameReport()
        self.setup()
    
    def teardown(self): 
        # TODO: While the active competition is running we can't have this raising an Error but have it raise an error afterwards to be more robust 
        pass

    def get_device_id(self):
        return uuid.UUID(int=uuid.getnode()).hex[-12:]

    def connect(self, ip='localhost', port=1234):
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.connect((ip, port))
        try:
            self.name = str(self.name)
        except:
            raise Exception("Agent must have a Stringifiable Name")
        self.respond_to_request("device_id", self.get_device_id())
        self.respond_to_request("name", self.name)
        data = self.client.recv(1024).decode()
        if data:
            response = json.loads(data)
            if response['message'] == 'provide_name':
                self.name = response['name']
                message = {
                    "message": "name_updated",
                }
                self.client.send(json.dumps(message).encode())
                print(f"My name is {self.name}")
        self.play()

    def get_action(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError

    def play(self):
        raise NotImplementedError

    def close(self):
        self.client.close()



from agt_server.local_games.base import LocalArena
from itertools import permutations
import numpy as np


class BOSArena(LocalArena):
    def __init__(self, num_rounds=1000, players=[], timeout=1, handin=False, save_path = None):
        super().__init__(num_rounds, players, timeout, handin, save_path)
        self.game_name = "Battle of the Sexes"
        self.valid_actions = [0, 1]
        self.utils = [[(0, 0), (3, 7)],
                      [(7, 3), (0, 0)]]
        self.invalid_move_penalty = -1
        if not self.handin_mode:
            assert len(self.players) >= 2, "Arena must have at least 2 players"

        for idx in range(len(self.players)):
            if self.handin_mode:
                try:
                    player = self.players[idx]
                    self.game_reports[player.name] = {
                        "action_history": [],
                        "util_history": [],
                        "index": idx,
                        "timeout_count": 0,
                        "global_timeout_count": 0,
                        "disconnected": False
                    }
                except:
                    continue
            else:
                player = self.players[idx]
                self.game_reports[player.name] = {
                    "action_history": [],
                    "util_history": [],
                    "index": idx,
                    "global_timeout_count": 0
                }

        import numpy as np
        self.result_table = np.zeros([len(players), len(players)])
        self.game_num = 1

    def reset_game_reports(self):
        for player in self.players:
            if self.handin_mode:
                try:
                    self.game_reports[player.name]["action_history"] = []
                    self.game_reports[player.name]["util_history"] = []
                except:
                    continue
            else:
                self.game_reports[player.name]["action_history"] = []
                self.game_reports[player.name]["util_history"] = []

    def run(self):
        for p1, p2 in permutations(self.players, 2):
            if self.handin_mode:
                if p1 is None or p2 is None or self.game_reports[p1.name]['disconnected'] or self.game_reports[p2.name]['disconnected']:
                    continue
                else:
                    try:
                        self.run_func_w_time(p1.setup, self.timeout, p1.name)
                    except:
                        self.game_reports[p1.name]['disconnected'] = True
                        continue

                    try:
                        self.run_func_w_time(p2.setup, self.timeout, p2.name)
                    except:
                        self.game_reports[p2.name]['disconnected'] = True
                        continue

                    try:
                        self.run_game(p1, p2)
                    except:
                        self.reset_game_reports()
                        continue
            else:
                self.run_func_w_time(p1.setup, self.timeout, p1.name)
                self.run_func_w_time(p2.setup, self.timeout, p2.name)
                self.run_game(p1, p2)
        results = self.summarize_results()
        return results

    def calculate_utils(self, p1_action, p2_action):
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            utils = [0, 0]
        elif p1_action not in self.valid_actions:
            utils = [self.invalid_move_penalty, 0]
        elif p2_action not in self.valid_actions:
            utils = [0, self.invalid_move_penalty]
        else:
            utils = [self.utils[p1_action][p2_action][0],
                     self.utils[p1_action][p2_action][1]]
        return utils

    def run_game(self, p1, p2):
        p1.player_type = "bos_player"
        p2.player_type = "bos_player"
        for _ in range(self.num_rounds):
            if self.handin_mode:
                if self.game_reports[p1.name]['disconnected'] or self.game_reports[p2.name]['disconnected']:
                    break
                if self.game_reports[p1.name]['timeout_count'] < self.timeout_tolerance:
                    try:
                        p1_action = self.run_func_w_time(
                            p1.get_action, self.timeout, p1.name, -1)
                    except:
                        self.game_reports[p1.name]['disconnected'] = True
                        p1_action = -1
                else:
                    self.game_reports[p1.name]['disconnected'] = True
                    p1_action = -1

                if self.game_reports[p2.name]['timeout_count'] < self.timeout_tolerance:
                    try:
                        p2_action = self.run_func_w_time(
                            p2.get_action, self.timeout, p2.name, -1)
                    except:
                        self.game_reports[p2.name]['disconnected'] = True
                        p2_action = -1
                else:
                    self.game_reports[p2.name]['disconnected'] = True
                    p2_action = -1

            else:
                p1_action = self.run_func_w_time(
                    p1.get_action, self.timeout, p1.name, -1)
                p2_action = self.run_func_w_time(
                    p2.get_action, self.timeout, p2.name, -1)
            
            if p1_action not in self.valid_actions: 
                p1_action = -1
            if p2_action not in self.valid_actions: 
                p2_action = -1

            self.game_reports[p1.name]['action_history'].append(p1_action)
            self.game_reports[p2.name]['action_history'].append(p2_action)

            p1_util, p2_util = self.calculate_utils(p1_action, p2_action)

            self.game_reports[p1.name]['util_history'].append(p1_util)
            self.game_reports[p2.name]['util_history'].append(p2_util)

            p1.game_report.game_history['my_action_history'].append(p1_action)
            p1.game_report.game_history['my_utils_history'].append(p1_util)
            p1.game_report.game_history['opp_action_history'].append(p2_action)
            p1.game_report.game_history['opp_utils_history'].append(p2_util)

            p2.game_report.game_history['my_action_history'].append(p2_action)
            p2.game_report.game_history['my_utils_history'].append(p2_util)
            p2.game_report.game_history['opp_action_history'].append(p1_action)
            p2.game_report.game_history['opp_utils_history'].append(p1_util)

            if self.handin_mode:
                try:
                    self.run_func_w_time(p1.update, self.timeout, p1.name)
                except:
                    self.game_reports[p1.name]['disconnected'] = True

                try:
                    self.run_func_w_time(p2.update, self.timeout, p2.name)
                except:
                    self.game_reports[p2.name]['disconnected'] = True
            else:
                self.run_func_w_time(p1.update, self.timeout, p1.name)
                self.run_func_w_time(p2.update, self.timeout, p2.name)

        if not self.handin_mode:
            print(f"Game {self.game_num}:")
        for i in range(2):
            p = [p1, p2][i]
            op = [p2, p1][i]
            action_counts = [0, 0, 0]
            for action in self.game_reports[p.name]['action_history']:
                if action in [0, 1]:
                    action_counts[action] += 1
                else:
                    action_counts[2] += 1
            if not self.handin_mode:
                print(
                    f"{p.name} was COOPERATIVE {action_counts[0]} times and STUBBORN {action_counts[1]} times")
                if action_counts[2] > 0:
                    print(f"{p.name} submitted {action_counts[2]} invalid moves")
                if self.game_reports[p.name]['global_timeout_count'] > 0: 
                    print(f"{p.name} timed out {self.game_reports[p.name]['global_timeout_count']} times")
                    self.game_reports[p.name]['global_timeout_count'] = 0
            
            total_util = sum(self.game_reports[p.name]['util_history'])

            avg_util = total_util / \
                len(self.game_reports[p.name]['util_history'])
            if not self.handin_mode:
                print(
                    f"{p.name} got a total utility of {total_util} and a average utility of {avg_util}")
            self.result_table[self.game_reports[p.name]['index'],
                              self.game_reports[op.name]['index']] += total_util
        self.game_num += 1
        self.reset_game_reports()

    def summarize_results(self):
        import pandas as pd
        df = pd.DataFrame(self.result_table)
        agent_names = [player.name for player in self.players if player is not None]
        df.columns = agent_names
        df.index = agent_names
        means = []

        for pld, d in zip(self.game_reports.values(), self.result_table):
            if self.handin_mode and pld['disconnected']:
                means.append(float('-inf'))
            else:
                means.append(sum(d) / (len(self.result_table) - 1))

        df['Mean Points'] = means
        final_scores = [m / self.num_rounds for m in means]
        df['Final Score'] = final_scores

        df = df.sort_values('Mean Points', ascending=False)

        df = df[df['Mean Points'] != float('-inf')]

        if not self.handin_mode:
            print(df)
        else: 
            np.savetxt(self.save_path, self.result_table, fmt='%d')
        return df



import threading
import traceback
from datetime import datetime
import os
import json

class LocalArena:
    def __init__(self, num_rounds, players, timeout, handin, logging_path, save_path):
        self.num_rounds = num_rounds
        self.players = players
        self.timeout = timeout
        self.handin_mode = handin
        self.timeout_tolerance = 5
        self.game_reports = {}
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.logging_path = f"{logging_path}/{timestamp}_log.txt"
        self.shortcut_path = f"{logging_path}/shortcut.json"
        if self.handin_mode and not os.path.exists(self.shortcut_path):
            empty_dict = {}
            with open(self.shortcut_path, 'w') as file:
                json.dump(empty_dict, file)
        
        self.save_path = save_path

    def run_func_w_time(self, func, timeout, name, alt_ret=None):
        def target_wrapper():
            nonlocal ret
            try:
                ret = func()
            except Exception as e:
                stack_trace = traceback.format_exc()
                if self.handin_mode: 
                    self.game_reports[name]['disconnected'] = True
                    with open(self.logging_path, 'a') as file:
                        file.write(f"Exception in thread running {name}: {e}\nStack Trace:\n{stack_trace}\n")
                else: 
                    print(f"Exception in thread running {name}: {e}\nStack Trace:\n{stack_trace}")
        
        thread = threading.Thread(target=target_wrapper)
        thread.start()
        thread.join(timeout)

        if thread.is_alive():
            thread.join()
            if not self.handin_mode:
                print(f"{name} Timed Out")
            else:
                with open(self.logging_path, 'a') as file:
                    file.write(f"{name} Timed Out\n")
            
            if name in self.game_reports:
                if 'timeout_count' in self.game_reports[name]:
                    self.game_reports[name]['timeout_count'] += 1
                if 'global_timeout_count' in self.game_reports[name]:
                    self.game_reports[name]['global_timeout_count'] += 1
        if ret is None: 
            ret = alt_ret
        
        return ret

    def run_game(self):
        raise NotImplementedError
